% =========================================================================
%
% The example is provided as a demonstration in:
% G. Molnar and N. Blala, Topology optimization of
%    periodic beam lattices using Cosserat elasticity
%
% to demonstrates how the topology optimization of
% a Cosserat medium is carried out
%
% =========================================================================
%
close all
clear all
hold off
clc
% =========================================================================
% ------------- MODEL -------------
% =========================================================================
% Input parameters
nely=75;           % Number of finite elements in the Y direction
LRat=2.;            % Aspect ratio of the design space
nelx=nely*LRat;     % Number of finite element in the X direction
volfrac=0.075;      % Global volume fraction
fRVEmax=0.4;       % Maximum local volume fraction

cnt=1;
% Generating finite element mesh
% Nodes
for j=0:nely
    for i=0:nelx
        coor(cnt,:)=[i j];
        cnt=cnt+1;
    end
end
Lmax=max(coor(:,2));
coor(:,1)=coor(:,1)/Lmax*50;
coor(:,2)=coor(:,2)/Lmax*50;
Left=min(coor(:,1)); Right=max(coor(:,1)); Btm=min(coor(:,2)); Top=max(coor(:,2));
% Elements
cnt=1;
for j=1:nely
    for i=1:nelx
        connec(cnt,:)=[(j-1)*(nelx+1)+i (j-1)*(nelx+1)+i+1  (j)*(nelx+1)+i+1 (j)*(nelx+1)+i];
        cnt=cnt+1;
    end
end
% Applying boundary conditions
sup=find(coor(:,1)<(Left+1e-3));
BC.sup=[sup*3-2;sup*3-1;sup*3]';
BC.dispbond=zeros(length(BC.sup),1);
btmNodes=find((coor(:,2)<(Btm+1e-4)));
bin=find(coor(btmNodes,1)>=(Right-1-1e-9));
BC.supF=btmNodes(bin)*3-1;
if bin>1
    BC.Fval=zeros(length(BC.supF),1)-1e-3/(length(bin)-1);
else
    BC.Fval=zeros(length(BC.supF),1)-1e-3;
end
BC.Fval(1)=BC.Fval(1)/2;
BC.Fval(end)=BC.Fval(end)/2;
% Material properties
Mat.E = 1; 
Mat.lmicro=100/300;
Mat.type='Square';

% Stacking model to structure
Model.coor=coor;
Model.connec=connec;
Model.BC=BC;
Model.Mat=Mat;

% =========================================================================
% ------------- OPTIMIZATION SETTINGS -------------
% =========================================================================
Opt.xfull=Mat.lmicro/2;
Opt.xmax=Mat.lmicro/2*fRVEmax;                      % For rectangular grid
%Opt.xmax=Mat.lmicro*fRVEmax*sqrt(3)/2;             % For honeycombs
%Opt.xmax=Mat.lmicro*fRVEmax*sqrt(3)/(2+2*sqrt(3)); % For reinforced honeycombs

Opt.rmin=1;
Opt.xmin=Opt.xmax/5000;
Opt.name=['CosseratSquare_rho'  num2str(volfrac) '_nx' num2str(nelx) '_fmax' num2str(fRVEmax)];
Opt.ccntmax=10;
Opt.dxmin=(Opt.xmax-Opt.xmin)/1000;
Opt.loopmax=500;

% =========================================================================
% ------------- OPTIMIZATION -------------
% =========================================================================
x=topOptimCosserat(Model,volfrac,Opt);

% =========================================================================
% ------------- DISPLAYING RESULTS -------------
% =========================================================================
cnt=1;
xp=zeros(nely,nelx);
for j=1:nely
    for i=1:nelx
        xp(nely-j+1,i)=x(cnt);
        cnt=cnt+1;
    end
end
colormap(gray); imagesc(-xp);
axis equal; axis tight; axis off; box on;

