% ======================================================================
% ----- Topology optimization algorythm with 2D Cosserat FEM ---- %
% ======================================================================
function x=topOptimCosserat(Model,volfrac,Opt)
% Input parameters
%   Model.coor - coordinates of nodes,
%   Model.connec - connectivity table of the 4-node rectangular elements,
%   Model.BC - boundary conditions,
%   Model.Mat - Material properties,
%   volfrac - volmue fraction,
%   Opt.rmin - filter distance,
%   Opt.name - name of the optimization,
%   Opt.ccntmax - Maximum number of iteration without a better structure,
%   Opt.dxmin - Minimum size of volume ratio change,
%   Opt.loopmax - Maximum number of iteration.
%
% Output:
%   x - a connec length array containing the design variables.
%
% The algorithm is based on the work of:
%     O. Sigmund, A 99 line topology optimization code written in matlab,
%       Structural and multidisciplinary optimization 21 (2) (2001) 120–127.
% 
% ======================================================================
%
% Copyright© Gergely Molnar, CNRS, INSA of Lyon, LaMCoS, UMR5259
%
% gergely.molnar@insa-lyon.fr
% http://www.molnar-research.com
%
% This software is a computer program whose purpose is to calculate
% maximum stress as a function of the phase-field parameters and
% the material's stiffness
%
% This software is governed by the CeCILL-B license under French law and
% abiding by the rules of distribution of free software. You can  use, 
% modify and/or redistribute the software under the terms of the CeCILL-B
% license as circulated by CEA, CNRS and INRIA at the following URL
% "http://www.cecill.info". 
%
% As a counterpart to the access to the source code and  rights to copy,
% modify and redistribute granted by the license, users are provided only
% with a limited warranty  and the software's author,  the holder of the
% economi% rights,  and the successive licensors  have only  limited
% liability. 
%
% In this respect, the user's attention is drawn to the risks associated
% with loading,  using,  modifying and/or developing or reproducing the
% software by the user in light of its specifi% status of free software,
% that may mean  that it is complicated to manipulate,  and  that  also
% therefore means  that it is reserved for developers  and  experienced
% professionals having in-depth computer knowledge. Users are therefore
% encouraged to load and test the software's suitability as regards their
% requirements in conditions enabling the security of their systems and/or 
% data to be ensured and,  more generally, to use and operate it in the 
% same conditions as regards security. 
%
% The fact that you are presently reading this means that you have had
% knowledge of the CeCILL-B
%
% ======================================================================
disp('------------- Optimization started -------------')
% ----------------- Loading input variables ----------------------------
coor=Model.coor;
connec=Model.connec;
BC=Model.BC;
Mat=Model.Mat;
rmin=Opt.rmin;
name=Opt.name;
loopmax=Opt.loopmax;
ccntmax=Opt.ccntmax;
dxmin=Opt.dxmin;
xmin=Opt.xmin;
xmax=Opt.xmax;
xfull=Opt.xfull;

% ------------- Initial volume fraction/beam height ----------------
x=zeros(length(connec(:,1)),1)+xfull*volfrac;

% ------------- Element centers and element neighbours -------------
nnodee=length(connec(1,:));
nelem=length(connec(:,1));
XC=zeros(length(connec(:,1)),2,nnodee);
for i=1:nnodee
XC(:,:,i)=[coor(connec(:,i),1) coor(connec(:,i),2)];
end
elemCenter=mean(XC,3);
for i=1:nelem
    distElem=sum((repmat(elemCenter(i,:),nelem,1)-elemCenter).^2,2).^0.5;
    elemNeigh{i}=find(distElem<=rmin);
end
elemWeight=polyarea(squeeze(XC(:,1,:))',squeeze(XC(:,2,:))')'/xfull;
voltot=sum(x.*elemWeight);

% ------------- Initialize -------------
c0=9e99;
toler=1e-6;
loop = 0; 
change = 1.;
ccnt=0;

% ======================================================================
% -------------------------- Start iterating ---------------------------
% ======================================================================
while change > dxmin && ccnt<ccntmax && loop<loopmax 
  loop = loop + 1;
  xold = x;
  
% --------------------- Mechanical equilibrium -------------------------
[U]=FEM2DCosserat(coor,connec,Mat,x,BC);

% ------------- Objective function and sensitivity analysis ------------
% OBJECTIVE FUNCTION AND SENSITIVITY ANALYSIS
[c,dc]=Energy2DCosserat(coor,connec,U,Mat,x);

% -------------------------- Filtering ---------------------------------
if rmin>0; [dc]   = checkMG(elemCenter,elemNeigh,rmin,x,dc); end
% ----------- Update based on the optimality criterion -----------------
[x]    = OC(voltot,elemWeight,x,dc,xmin,xmax);

% --------------------- Checking convergence ---------------------------
% --- Local variable change ---
change = max(max(abs(x-xold)));
% --- Energy change ---
if loop>1
    cch=c-min(cvect);
%    cch=c-max(cvect);  % for displacement optimization
    if cch>toler
%    if cch<toler % for displacement optimization
       ccnt=ccnt+1;
    else
       ccnt=0;
    end
end
% --- Displacement at boundary ---
uMax=max(U(BC.supF));

% --- Display actual state of optimization ----
disp([' It.: ' sprintf('%4i',loop) ' Obj.: ' sprintf('%10.4g',c) ' Opt.: ' sprintf('%4i',ccnt)  ...
      ' Mass: ' sprintf('%10.4g',sum(x.*elemWeight)) ' Max U: ' sprintf('%10.4g',uMax) ...
      ' ch.: ' sprintf('%10.4g',change )])

% --------------------- Store results ---------------------------
cvect(loop)=c;
changevect(loop)=change;
uMaxvect(loop)=uMax;
save(['OptimTopo_' name '.mat'],'connec','coor','x','cvect','U','BC','elemWeight','elemCenter','rmin','changevect','uMaxvect','ccnt','Mat')

  
end
if ccnt>=ccntmax
    disp('Energy precision is exceeded')
else
    disp('Further change is small')
end
disp('------------- Optimization ended -------------')
end
% ======================================================================
% ------------------- Cosserat theory based FEM code -------------------
% ======================================================================
function [U]=FEM2DCosserat(coor,connec,Mat,x,BC)

sup=BC.sup;
dispbond=BC.dispbond;
supF=BC.supF;
Fval=BC.Fval;
E=Mat.E;
lmicro=Mat.lmicro;
type=Mat.type;

nelem=length(connec(:,1));
nnode=length(coor(:,1));

% --- Initializing main vectors and matrices ---
K = sparse(3*nnode, 3*nnode);
F = zeros(3*nnode,1); U = zeros(3*nnode,1);
Fext = sparse(3*nnode,1);

% --- Local integration point coordinates ---
sqr3=1/sqrt(3);
XII=[-sqr3 -sqr3;sqr3 -sqr3;sqr3 sqr3;-sqr3 sqr3];
% Coordinates of each nodes in elements
X1=[coor(connec(:,1),1) coor(connec(:,1),2)];
X2=[coor(connec(:,2),1) coor(connec(:,2),2)];
X3=[coor(connec(:,3),1) coor(connec(:,3),2)];
X4=[coor(connec(:,4),1) coor(connec(:,4),2)];

% --- Positions of the stifffness components ---
Ivect=zeros(nelem*12*12,1);
Jvect=zeros(nelem*12*12,1);
Kvect=zeros(nelem*12*12,1);

cntK=1;
for e=1:nelem
    Kloc=zeros(12,12);
    PosLoc=[connec(e,:)];
    ind=1+3*(connec(e,:)-1);
    DOFLoc=[ind(1):ind(1)+2,ind(2):ind(2)+2,ind(3):ind(3)+2,ind(4):ind(4)+2];     
    COORDS=[X1(e,1:2);X2(e,1:2);X3(e,1:2);X4(e,1:2)]';        
% --- Loading materials stiffness matrix ---
    if contains(type,'Square')
        CMAT=cmatSquare(E,x(e),lmicro);
    elseif contains(type,'Honey') && contains(type,'Re')==0
        CMAT=cmatHoney(E,x(e),lmicro);
    elseif contains(type,'ReHoney')
        CMAT=cmatReHoney(E,x(e),lmicro);
    else
        erro('No microstructure defined')
    end
    for ipoint=1:4
        XI=XII(ipoint,:);
        AN =[1/4*(1-XI(1))*(1-XI(2)), 1/4*(1+XI(1))*(1-XI(2)), 1/4*(1+XI(1))*(1+XI(2)), 1/4*(1-XI(1))*(1+XI(2))];
        dNdxi=[-1/4*(1-XI(2)), -1/4*(1-XI(1));
               1/4*(1-XI(2)), -1/4*(1+XI(1));
               1/4*(1+XI(2)), 1/4*(1+XI(1));
               -1/4*(1+XI(2)), 1/4*(1-XI(1))];
        JACOB=COORDS*dNdxi;
        DTM=det(JACOB);
        dNdx=dNdxi/JACOB;

        BB=[dNdx(1,1) 0          0        dNdx(2,1) 0          0          dNdx(3,1) 0          0        dNdx(4,1) 0          0;
            0         dNdx(1,2)  0        0         dNdx(2,2)  0          0         dNdx(3,2)  0        0         dNdx(4,2)  0;
            dNdx(1,2) dNdx(1,1)  0        dNdx(2,2) dNdx(2,1)  0          dNdx(3,2) dNdx(3,1)  0        dNdx(4,2) dNdx(4,1)  0;
            dNdx(1,2) -dNdx(1,1) -2*AN(1) dNdx(2,2) -dNdx(2,1) -2*AN(2)   dNdx(3,2) -dNdx(3,1) -2*AN(3) dNdx(4,2) -dNdx(4,1) -2*AN(4);
            0         0          dNdx(1,1) 0         0          dNdx(2,1) 0         0          dNdx(3,1) 0         0          dNdx(4,1);
            0         0          dNdx(1,2) 0         0          dNdx(2,2) 0         0          dNdx(3,2) 0         0          dNdx(4,2)];

% --- Elementary stiffness matrix ---
        Kloc=Kloc+BB'*CMAT*BB*DTM;
    end
% --- Complition ---
     for i=1:12
          I=DOFLoc(i);
          for j=1:12
              J=DOFLoc(j);
              Ivect(cntK)=I;
              Jvect(cntK)=J;
              Kvect(cntK)=Kloc(i,j);
              cntK=cntK+1;
          end
      end
    
end
K=sparse(Ivect,Jvect,Kvect);

% --- Dirichlet boundary conditions ---
for i=1:length(sup)
   J=sup(i);
   F=F+(-1)*K(:,J)*dispbond(i);
end
F(sup)=-(-1)*dispbond;
K(sup,:)=0;
K(:,sup)=0;
for i=1:length(sup)
    K(sup(i),sup(i))=1;
end

% --- Neumann boundary conditions ---
Fext(supF)=Fval;
F=Fext-F;

% --- Solving the minear equation system ---
U = full(K\F);  


end
% ======================================================================
% -------- Calculating objective function and sensitivity array --------
% ======================================================================
function [c,dc]=Energy2DCosserat(coor,connec,U,Mat,x)

E=Mat.E;
lmicro=Mat.lmicro;
type=Mat.type;

nelem=length(connec(:,1));
nnode=length(coor(:,1));

% --- Local integration point coordinates ---
sqr3=1/sqrt(3);
XII=[-sqr3 -sqr3;sqr3 -sqr3;sqr3 sqr3;-sqr3 sqr3];

% --- Coordinates of each nodes in elements ---
X1=[coor(connec(:,1),1) coor(connec(:,1),2)];
X2=[coor(connec(:,2),1) coor(connec(:,2),2)];
X3=[coor(connec(:,3),1) coor(connec(:,3),2)];
X4=[coor(connec(:,4),1) coor(connec(:,4),2)];

c=0.;  
dc=zeros(nelem,1);
for e=1:nelem
    ind=1+3*(connec(e,:)-1);
    DOFLoc=[ind(1):ind(1)+2,ind(2):ind(2)+2,ind(3):ind(3)+2,ind(4):ind(4)+2];     
    Uloc=U(DOFLoc);
    COORDS=[X1(e,1:2);X2(e,1:2);X3(e,1:2);X4(e,1:2)]';        
% --- Loading elementary stiffness matrix ---
    if contains(type,'Square')
        CMAT=cmatSquare(E,x(e),lmicro);
        dCMAT=dcmatSquare(E,x(e),lmicro);
    elseif contains(type,'Honey') && contains(type,'Re')==0
        CMAT=cmatHoney(E,x(e),lmicro);
        dCMAT=dcmatHoney(E,x(e),lmicro);
    elseif contains(type,'ReHoney')
        CMAT=cmatReHoney(E,x(e),lmicro);
        dCMAT=dcmatReHoney(E,x(e),lmicro);
    else
        erro('No microstructure defined')
    end    
    for ipoint=1:4
        XI=XII(ipoint,:);
        AN =[1/4*(1-XI(1))*(1-XI(2)), 1/4*(1+XI(1))*(1-XI(2)), 1/4*(1+XI(1))*(1+XI(2)), 1/4*(1-XI(1))*(1+XI(2))];
        dNdxi=[-1/4*(1-XI(2)), -1/4*(1-XI(1));
               1/4*(1-XI(2)), -1/4*(1+XI(1));
               1/4*(1+XI(2)), 1/4*(1+XI(1));
               -1/4*(1+XI(2)), 1/4*(1-XI(1))];
        JACOB=COORDS*dNdxi;
        DTM=det(JACOB);
        dNdx=dNdxi/JACOB;

        BB=[dNdx(1,1) 0          0        dNdx(2,1) 0          0          dNdx(3,1) 0          0        dNdx(4,1) 0          0;
            0         dNdx(1,2)  0        0         dNdx(2,2)  0          0         dNdx(3,2)  0        0         dNdx(4,2)  0;
            dNdx(1,2) dNdx(1,1)  0        dNdx(2,2) dNdx(2,1)  0          dNdx(3,2) dNdx(3,1)  0        dNdx(4,2) dNdx(4,1)  0;
            dNdx(1,2) -dNdx(1,1) -2*AN(1) dNdx(2,2) -dNdx(2,1) -2*AN(2)   dNdx(3,2) -dNdx(3,1) -2*AN(3) dNdx(4,2) -dNdx(4,1) -2*AN(4);
            0         0          dNdx(1,1) 0         0          dNdx(2,1) 0         0          dNdx(3,1) 0         0          dNdx(4,1);
            0         0          dNdx(1,2) 0         0          dNdx(2,2) 0         0          dNdx(3,2) 0         0          dNdx(4,2)];
        
% --- Total potential energy ---
        c=c+Uloc'*BB'*CMAT*BB*Uloc*DTM/2;
% --- Sensitivity array ---
        dc(e)=dc(e)+-Uloc'*BB'*dCMAT*BB*Uloc*DTM/2;
        
    end
end
end
% ======================================================================
% -------------------- Applying mesh-independency filter ---------------
% ======================================================================
function [dcn]=checkMG(elemCenter,elemNeigh,rmin,x,dc)
nelem=length(dc);
dcn=zeros(nelem,1);
for i=1:nelem
    nf=elemNeigh{i};
    distElem=sum([elemCenter(nf,1)-elemCenter(i,1) elemCenter(nf,2)-elemCenter(i,2)].^2,2).^0.5;
    Hf=rmin-distElem;
    xf=x(nf);
    dcdxf=dc(nf);
    dcn(i)=sum(Hf.*xf.*dcdxf)/(x(i)*sum(Hf));
end
end
% ======================================================================
% ----------------------- Updating optimality contdition ---------------
% ======================================================================
function [xnew]=OC(voltot,elemWeight,x,dc,xmin,xmax)
dc=dc/max(abs(dc));
move=(xmax-xmin)/2;
l1 = 0; l2 = 1000000;
while (l2-l1 > 1e-8)
  lmid = 0.5*(l2+l1);
  xnew = max(xmin,max(x-move,min(xmax,min(x+move,x.*sqrt(-dc./lmid)))));
  if sum(sum(xnew.*elemWeight)) - voltot > 0
    l1 = lmid;
  else
    l2 = lmid;
  end
end
end
% ======================================================================
% ----------------------- Material properties --------------------------
% ======================================================================
% ----------------------- Rectangular lattice --------------------------
function CMAT=cmatSquare(E,x,lmicro)
CMAT=E*[x/lmicro 0 0     0 0 0
        0 x/lmicro 0     0 0 0
        0 0 x^3/lmicro^3/2 0 0 0
        0 0 0 x^3/lmicro^3/2 0 0
        0 0 0 0 x^3/12/lmicro 0
        0 0 0 0 0 x^3/12/lmicro];
end
function dCMAT=dcmatSquare(E,x,lmicro)
dCMAT=E*[1/lmicro 0 0     0 0 0
        0 1/lmicro 0     0 0 0
        0 0 3*x^2/lmicro^3/2 0 0 0
        0 0 0 3*x^2/lmicro^3/2 0 0
        0 0 0 0 3*x^2/12/lmicro 0
        0 0 0 0 0 3*x^2/12/lmicro];
end
% ----------------------- Hexagonal lattice -----------------------------
function CMAT=cmatHoney(E,x,lmicro)
Ks=sqrt(3)*E*x/6/lmicro;
Gs=sqrt(3)*E*x^3/(3*lmicro*(lmicro^2+x^2));
Gc=sqrt(3)*E*x^3/6/lmicro^3;
gamma=E*x^3/12/lmicro/sqrt(3);
CMAT=[Ks+Gs Ks-Gs 0 0 0 0
        Ks-Gs Ks+Gs 0 0 0 0
        0 0  Gs 0 0 0
        0 0 0 Gc 0 0
        0 0 0 0 gamma 0
        0 0 0 0 0 gamma];
end
function dCMAT=dcmatHoney(E,x,lmicro)
dKs=sqrt(3)*E/6/lmicro;
dGs=sqrt(3)*E/(3*lmicro)*((3*x^2)/(x^2+lmicro^2)-(2*x^4)/(x^2+lmicro^2)^2);
dGc=sqrt(3)*E*3*x^2/6/lmicro^3;
dgamma=E*3*x^2/12/lmicro/sqrt(3);
dCMAT=[dKs+dGs dKs-dGs 0 0 0 0
        dKs-dGs dKs+dGs 0 0 0 0
        0 0  dGs 0 0 0
        0 0 0 dGc 0 0
        0 0 0 0 dgamma 0
        0 0 0 0 0 dgamma];
end
% ---------------- Reinforced hexagonal lattice -------------------------
function CMAT=cmatReHoney(E,x,LGrid)
a=1;
I=a*x^3/12;
S=a*x;
Ks=(3+sqrt(3))/6*E*S/LGrid;
Gs=(48*I^2+16*(sqrt(3)+1)*LGrid^2*I*S+LGrid^4*S^2)/(4*LGrid^3*(LGrid^2*S+12*I))*E;
Gc=(3+sqrt(3))/3*2*E*I/LGrid^3*sqrt(3);
gamma=(1+sqrt(3))*sqrt(3)/3*I*E/LGrid;
CMAT=[Ks+Gs Ks-Gs 0 0 0 0
        Ks-Gs Ks+Gs 0 0 0 0
        0 0  Gs 0 0 0
        0 0 0 Gc 0 0
        0 0 0 0 gamma 0
        0 0 0 0 0 gamma];
end
function dCMAT=dcmatReHoney(E,x,LGrid)
a=1;
I=a*x^3/12;
S=a*x;
dI=a*x^2/4;
dS=a;
dKs=(3+sqrt(3))/6*E/LGrid*a;
gx=(48*I^2+16*(sqrt(3)+1)*LGrid^2*I*S+LGrid^4*S^2);
hx=(4*LGrid^3*(LGrid^2*S+12*I));
dgx=96*I*dI+16*(sqrt(3)+1)*LGrid^2*(dI*S+dS*I)+LGrid^4*2*S*dS;
dhx=(4*LGrid^3*(LGrid^2*dS+12*dI));
dGs=E*(dgx*hx-dhx*gx)/hx^2;
dGc=(3+sqrt(3))/3*2*E*dI/LGrid^3*sqrt(3);
dgamma=(1+sqrt(3))*sqrt(3)/3*dI*E/LGrid;
dCMAT=[dKs+dGs dKs-dGs 0 0 0 0
        dKs-dGs dKs+dGs 0 0 0 0
        0 0  dGs 0 0 0
        0 0 0 dGc 0 0
        0 0 0 0 dgamma 0
        0 0 0 0 0 dgamma];
end

